package jesse.keeblarcraft.CustomBlocks.BlockEntities;

import java.util.ArrayList;

import jesse.keeblarcraft.AttributeMgr.AttributeMgr;
import jesse.keeblarcraft.AttributeMgr.AttributeTree;
import jesse.keeblarcraft.FactionMgr.Callbacks.PlayerEnteredBaseCallback;
import jesse.keeblarcraft.FactionMgr.Callbacks.PlayerExitedBaseCallback;
import jesse.keeblarcraft.FactionMgr.Callbacks.PlayerInBaseCallback;
import jesse.keeblarcraft.GuiMgr.FactionBlockScreenHandler;
import jesse.keeblarcraft.world.ImplementedInventory;
import net.fabricmc.fabric.api.screenhandler.v1.ExtendedScreenHandlerFactory;
import net.minecraft.block.BlockState;
import net.minecraft.block.entity.BlockEntity;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.entity.player.PlayerInventory;
import net.minecraft.inventory.Inventories;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NbtCompound;
import net.minecraft.network.PacketByteBuf;
import net.minecraft.screen.PropertyDelegate;
import net.minecraft.screen.ScreenHandler;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.text.Text;
import net.minecraft.util.ActionResult;
import net.minecraft.util.collection.DefaultedList;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;

public class FactionBlockEntity extends BlockEntity implements ExtendedScreenHandlerFactory, ImplementedInventory {
    private final DefaultedList<ItemStack> inventory = DefaultedList.ofSize(7, ItemStack.EMPTY);

    private static final int DEFENSE_SLOT_ONE = 0;
    private static final int DEFENSE_SLOT_TWO = 1;
    private static final int OFFENSE_SLOT_ONE = 2;
    private static final int OFFENSE_SLOT_TWO = 3;
    private static int factionPower = 0;
    Boolean stopMobSpawn = true;
    Boolean hasBuildFlight = true;
    private ArrayList<String> playersInRadius = new ArrayList<>();

    protected final PropertyDelegate propertyDelegate;

    public FactionBlockEntity(BlockPos pos, BlockState state) {
        super(BlockEntityRegistration.FACTION_BLOCK_ENTITY, pos, state);
        this.propertyDelegate = new PropertyDelegate() {

            @Override
            public int get(int index) {
                // The only value we need to get/delegate is faction power
                return switch(index) {
                    default -> factionPower;
                };
            }

            @Override
            public void set(int index, int value) {
                switch(index) {
                    default -> factionPower = value;
                };
            }

            @Override
            public int size() {
                // We are only sync'ing 1 integer - faction power
                return 1;
            }

        };
    }

    @Override
    public Text getDisplayName() {
        return Text.literal("Faction Home Base Station"); // Replace with proper en_us format later
    }

    @Override
    public ScreenHandler createMenu(int syncId, PlayerInventory playerInventory, PlayerEntity player) {
        return new FactionBlockScreenHandler(syncId, playerInventory, this, this.propertyDelegate);
    }

    @Override
    public void readNbt(NbtCompound nbt) {
        super.readNbt(nbt);
        Inventories.readNbt(nbt, inventory);
        factionPower = nbt.getInt("faction_block_entity.power");
    }

    @Override
    public void writeNbt(NbtCompound nbt) {
        super.writeNbt(nbt);
        // Write the inventory of the block
        Inventories.writeNbt(nbt, inventory);
        nbt.putInt("faction_block_entity.power", factionPower);
    }

    @Override
    public DefaultedList<ItemStack> getItems() {
        return inventory;
    }

    @Override
    public void writeScreenOpeningData(ServerPlayerEntity player, PacketByteBuf buf) {
        // Send position of this block entity from server->client->server etc
        buf.writeBlockPos(this.pos);
    }

    private Boolean IsPlayerInBounds(PlayerEntity player, BlockPos pos) {
        Boolean isNearBlock = false;

        // Will check in range - assumes same world! 50 is a temporary value at the moment
        Boolean xBounds = player.getX() <= pos.getX() + 50 && player.getX() >= pos.getX() - 50;
        Boolean yBounds = player.getY() <= pos.getY() + 50 && player.getY() >= pos.getY() - 50;
        Boolean zBounds = player.getZ() <= pos.getZ() + 50 && player.getZ() >= pos.getZ() - 50;
        if (xBounds && yBounds && zBounds) {
            isNearBlock = true;
        }

        return isNearBlock;
    }

    // Tick method is called 20 times a second
    public void tick(World world, BlockPos pos, BlockState state) {
        // For reasons unknown to me and only to Kaupenjoe (youtube video) - we never want to call this on a client.
        if (world.isClient()) {
            return;
        }

        // TODO: Optimize this block so that when it is placed the placers UUID is related to a faction and only pull from a list of those players not the entire server
        for (PlayerEntity player : world.getPlayers()) {
            Boolean isPlayerInFactionRadius = IsPlayerInBounds(player, pos);
            if (isPlayerInFactionRadius) {
                // Run individual faction modules for players here
                // First time entry callback check
                if (!playersInRadius.contains(player.getUuidAsString())) {
                    playersInRadius.add(player.getUuidAsString());
                    ActionResult result = PlayerEnteredBaseCallback.EVENT.invoker().interact(player, world);
                }

                // Invoke the flight attribute on this player
                if (hasBuildFlight) {
                    AttributeMgr.ApplyAttribute(player.getUuidAsString(), "faction_flight");
                    ActionResult result = PlayerInBaseCallback.EVENT.invoker().interact(player, world);
                }
            } else if (!isPlayerInFactionRadius && playersInRadius.contains(player.getUuidAsString())) {
                playersInRadius.remove(player.getUuidAsString());
                ActionResult result = PlayerExitedBaseCallback.EVENT.invoker().interact(player, world);
            }
        }

        // if (stopMobSpawn) {
        //     // Temporary for now
        //     // Sphere center point is going to be X,Y,Z
        //     Integer X = pos.getX();
        //     Integer y = pos.getY();
        //     Integer z = pos.getZ();
        //     Integer radius = 10;
        // }

        // if (hasBuildFlight) {
        //     for (PlayerEntity player : world.getPlayers()) {
        //         if (IsPlayerInBounds(player, pos)) {
        //             // Notify the attribute tree to enable this attribute!

        //         }
        //     }

        //     world.getBlockEntity(pos.east());
        // }
    }
}
